from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import os, math, sys, shutil, gzip
import casm.vasp
from casm.vasp import io


class Relax(object):
    """The Relax class contains functions for setting up, executing, and parsing a VASP relaxation.

        The relaxation is initialized in a directory containing VASP input files, called 'calcdir'.
        It then creates the following directory structure:
        .../calcdir/
            run.0/
            run.1/
            ...
            run.final/



        'run.i' directories are only created when ready.
        'run.final' is a final constant volume run {"ISIF":2, "ISMEAR":-5, "NSW":0, "IBRION":-1}.

        Contains:
            self.calcdir  (.../relax)
            self.rundir    (list of .../relax/run.i)
            self.finaldir (.../relax/run.final)
            self.subdir (adds to path of run.i for subdirectory runs)
    """
    def __init__(self, calcdir=None, settings=None):
        """
        Construct a VASP relaxation job object.

        Args:
            relaxdir:  path to vasp relaxation directory
            settings:   dictionary-like object containing settings, or if None, it reads
                        the json file: .../calcdir/relax.json

                possible settings keys are:
                    used by casm.vasp.run() function:
                        "ncpus": number of ncpus to run mpi on
                        "npar" or "ncore": number of ways to parallelize
                        "kpar": number of ways to parallelize k-points
                        "vasp_cmd": (default, see casm.vasp.run) shell command to execute vasp, or None to use default mpirun
                        "strict_kpoint": force strict copying of KPOINTS file, otherwise kpoints are scaled based on supercell size
                    used by not_converging():
                        "run_limit": (default 10) maximum number of runs to allow before setting status to "not_converging"

        """

        print("Constructing a VASP Relax object")
        sys.stdout.flush()
        self.subdir = ""
        # store path to .../calcdir, and create if not existing
        if calcdir is None:
            calcdir = os.getcwd()
        self.calcdir = os.path.abspath(calcdir)

        print("  Calculation directory:", self.calcdir)
        sys.stdout.flush()

        # find existing .../calcdir/run.run_index directories, store paths in self.rundir list
        self.rundir = []
        self.errdir = []
        self.update_rundir()
        self.update_errdir()

        self.finaldir = os.path.join(self.calcdir, "run.final")

        if settings is None:
            self.settings = dict()
        else:
            self.settings = settings

        # set default settings:
        if not "npar" in self.settings:
            self.settings["npar"] = None
        if not "kpar" in self.settings:
            self.settings["kpar"] = None
        if not "ncore" in self.settings:
            self.settings["ncore"] = None
        if not "vasp_cmd" in self.settings:
            self.settings["vasp_cmd"] = None
        if not "ncpus" in self.settings:
            self.settings["ncpus"] = None
        if not "run_limit" in self.settings:
            self.settings["run_limit"] = 10
        if not "nrg_convergence" in self.settings:
            self.settings["nrg_convergence"] = None
        if not "compress" in self.settings:
            self.settings["compress"] = []
        if not "move" in self.settings:
            self.settings["move"] = []
        if not "copy" in self.settings:
            self.settings["copy"] = []
        if not "backup" in self.settings:
            self.settings["backup"] = []
        if not "remove" in self.settings:
            self.settings["remove"] = []
        if not "extra_input_files" in self.settings:
            self.settings["extra_input_files"] = []
        if not "initial" in self.settings:
            self.settings["initial"] = None
        if not "final" in self.settings:
            self.settings["final"] = None
        if not "err_types" in self.settings:
            self.settings["err_types"] = ['SubSpaceMatrixError']
        if "subdir" in self.settings:
            self.subdir = self.settings["subdir"]

        print("VASP Relax object constructed\n")
        sys.stdout.flush()

    def add_rundir(self):
        """Make a new run.i directory"""
        os.makedirs(
            os.path.join(self.calcdir, "run." + str(len(self.rundir)),
                         self.subdir))
        self.update_rundir()
        self.update_errdir()

    def update_rundir(self):
        """Find all .../config/vasp/relax/run.i directories, store paths in self.rundir list"""
        self.rundir = []
        run_index = len(self.rundir)
        while os.path.isdir(
                os.path.join(self.calcdir, "run." + str(run_index),
                             self.subdir)):
            self.rundir.append(
                os.path.join(self.calcdir, "run." + str(run_index),
                             self.subdir))
            run_index += 1

    def add_errdir(self):
        """Move run.i to run.i_err.j directory"""
        os.rename(self.rundir[-1],
                  self.rundir[-1] + "_err." + str(len(self.errdir)))
        self.update_errdir()

    def update_errdir(self):
        """Find all .../config/vasp/relax/run.i_err.j directories, store paths in self.errdir list"""
        self.errdir = []
        if len(self.rundir) == 0:
            pass
        else:
            err_index = len(self.errdir)
            while os.path.isdir(self.rundir[-1] + "_err." + str(err_index)):
                self.errdir.append(self.rundir[-1] + "_err." + str(err_index))
                err_index += 1

    def setup(self, initdir, settings):
        """ mv all files and directories (besides initdir) into initdir """

        print("Moving files into initial run directory:", initdir)
        initdir = os.path.abspath(initdir)
        for p in os.listdir(self.calcdir):
            if (p in (io.VASP_INPUT_FILE_LIST +
                      self.settings["extra_input_files"])) and (os.path.join(
                          self.calcdir, p) != initdir):
                os.rename(os.path.join(self.calcdir, p),
                          os.path.join(initdir, p))
        print("")
        sys.stdout.flush()
        if "subdir" in settings:
            if settings["subdir"] == "01":
                new_values = {"IMAGES": None, "SPRING": None}
                io.set_incar_tag(new_values, initdir)
        # Keep a backup copy of the base INCAR
        shutil.copyfile(os.path.join(initdir, "INCAR"),
                        os.path.join(self.calcdir, "INCAR.base"))

        # If an initial incar is called for, copy it in and set the appropriate flag
        if (self.settings["initial"] != None) and (os.path.isfile(
                os.path.join(self.calcdir, self.settings["initial"]))):
            new_values = io.Incar(
                os.path.join(self.calcdir, self.settings["initial"])).tags
            io.set_incar_tag(new_values, initdir)
            print("  Set INCAR tags:", new_values, "\n")
            sys.stdout.flush()

    def complete(self):
        """Check if the VASP relaxation is complete.

           Completion criteria: .../config/vasp/relax/run.final/OUTCAR exists and is complete
        """
        outcarfile = os.path.join(self.finaldir, "OUTCAR")
        if not os.path.isfile(outcarfile):
            return False
        if not io.Outcar(outcarfile).complete():
            return False
        return True

    def converged(self):
        """Check if configuration is relaxed.

           This is called when self.rundir[-1] is complete and not a constant volume job.

           Convergence criteria is: at least 2 relaxation jobs are complete, and:
                                    1) the last job completed with <= 3 ionic steps
                                    or 2) the last two jobs had final E0 differ by less than
                                          self.settings["nrg_convergence"]
        """
        if len(self.rundir) >= 2:
            if io.ionic_steps(self.rundir[-1]) <= 3:
                return True
            if self.settings["nrg_convergence"] != None:
                if io.job_complete(self.rundir[-1]) and io.job_complete(
                        self.rundir[-2]):
                    o1 = io.Oszicar(os.path.join(self.rundir[-1], "OSZICAR"))
                    o2 = io.Oszicar(os.path.join(self.rundir[-2], "OSZICAR"))
                    if abs(o1.E[-1] -
                           o2.E[-1]) < self.settings["nrg_convergence"]:
                        return True

        return False

    def not_converging(self):
        """Check if configuration is not converging.

           This is called when self.rundir[-1] is complete and not a constant volume job and self.converged() == False.

           Not converging criteria: >= 10 runs without completion
        """
        if len(self.rundir) >= int(self.settings["run_limit"]):
            return True
        return False

    def run(self):
        """ Perform a series of vasp jobs to relax a structure. Performs a series of vasp calculations until
            convergence is reached according to the criteria in 'status()'. Then performs a final constant volume run
            {"ISIF":2, "ISMEAR":-5, "NSW":0, "IBRION":-1}.
        """

        print("Begin VASP relaxation run")
        sys.stdout.flush()

        # get current status of the relaxation:
        (status, task) = self.status()
        print("\n++  status:", status, "  next task:", task)
        sys.stdout.flush()

        while status == "incomplete":
            if task == "setup":
                self.add_rundir()
                self.setup(self.rundir[-1], self.settings)

            elif task == "relax":
                self.add_rundir()
                vasp.continue_job(self.rundir[-2], self.rundir[-1],
                                  self.settings)
                shutil.copyfile(os.path.join(self.calcdir, "INCAR.base"),
                                os.path.join(self.rundir[-1], "INCAR"))

            elif task == "constant":
                self.add_rundir()
                casm.vasp.continue_job(self.rundir[-2], self.rundir[-1],
                                       self.settings)

                # set INCAR to ISIF = 2, ISMEAR = -5, NSW = 0, IBRION = -1
                if (self.settings["final"] != None) and (os.path.isfile(
                        os.path.join(self.calcdir, self.settings["final"]))):
                    new_values = io.Incar(
                        os.path.join(self.calcdir,
                                     self.settings["final"])).tags
                else:
                    new_values = {
                        "ISIF": 2,
                        "ISMEAR": -5,
                        "NSW": 0,
                        "IBRION": -1
                    }

                # set INCAR system tag to denote 'final'
                if io.get_incar_tag("SYSTEM", self.rundir[-1]) is None:
                    new_values["SYSTEM"] = "final"
                else:
                    new_values["SYSTEM"] = io.get_incar_tag(
                        "SYSTEM", self.rundir[-1]) + " final"

                io.set_incar_tag(new_values, self.rundir[-1])
                print("  Set INCAR tags:", new_values, "\n")
                sys.stdout.flush()

            else:
                # probably hit walltime
                self.add_rundir()
                casm.vasp.continue_job(self.rundir[-2], self.rundir[-1],
                                       self.settings)

            while True:
                # run vasp
                result = casm.vasp.run(self.rundir[-1],
                                       npar=self.settings["npar"],
                                       ncore=self.settings["ncore"],
                                       command=self.settings["vasp_cmd"],
                                       ncpus=self.settings["ncpus"],
                                       kpar=self.settings["kpar"],
                                       err_types=self.settings["err_types"])

                # if no errors, continue
                if result is None or self.not_converging():
                    # Check for actions that should be taken after the initial run
                    if len(self.rundir) == 1:
                        if self.settings["fine_ngx"]:
                            outcarfile = os.path.join(self.rundir[-1],
                                                      "OUTCAR")
                            if not os.path.isfile(outcarfile):
                                # This is an error but I'm not sure what to do about it
                                pass
                            else:
                                init_outcar = io.Outcar(outcarfile)
                                if not init_outcar.complete:
                                    # This is an error but I'm not sure what to do about it
                                    pass
                                elif (init_outcar.ngx is None
                                      or init_outcar.ngy is None
                                      or init_outcar.ngz is None):
                                    # This is an error but I'm not sure what to do about it
                                    pass
                                else:
                                    ng_tags = {
                                        "ngx": init_outcar.ngx * 2,
                                        "ngy": init_outcar.ngy * 2,
                                        "ngz": init_outcar.ngz * 2
                                    }
                                    print(ng_tags)
                                    io.set_incar_tag(ng_tags, self.calcdir,
                                                     "INCAR.base")
                    break

                # else, attempt to fix first error
                self.add_errdir()
                os.mkdir(self.rundir[-1])
                # self.add_rundir()
                err = next(iter(result.values()))
                # err = result.itervalues().next()

                print("\n++  status:", "error", "  next task:", "fix_error")
                sys.stdout.flush()

                print("Attempting to fix error:", str(err))
                err.fix(self.errdir[-1], self.rundir[-1], self.settings)
                print("")
                sys.stdout.flush()

                if (self.settings["backup"] != None) and len(self.rundir) > 1:
                    print("Restoring from backups:")
                    for f in self.settings["backup"]:
                        if os.path.isfile(
                                os.path.join(self.rundir[-2],
                                             f + "_BACKUP.gz")):
                            f_in = gzip.open(
                                os.path.join(self.rundir[-2], f + "_BACKUP.gz",
                                             'rb'))
                            f_out = open(os.path.join(self.rundir[-1], f,
                                                      'wb'))
                            f_out.write(f_in.read())
                            f_in.close()
                            f_out.close()
                            print(f, " restored!")
                    sys.stdout.flush()

            (status, task) = self.status()
            print("\n++  status:", status, "  next task:", task)
            sys.stdout.flush()

        if status == "complete":
            if not os.path.isdir(self.finaldir):
                # mv final results to relax.final
                print("mv", os.path.basename(self.rundir[-1]),
                      os.path.basename(self.finaldir))
                sys.stdout.flush()
                os.rename(self.rundir[-1], self.finaldir)
                self.rundir.pop()
                casm.vasp.complete_job(self.finaldir, self.settings)

        return (status, task)

    def status(self):
        """ Determine the status of a vasp relaxation series of runs. Individual runs in the series
            are in directories labeled "run.0", "run.1", etc.

            Returns a tuple: (status = "incomplete" or "complete" or "not_converging",
                                task = continuedir or "relax" or "constant" or None)

            The first value is the status of the entire relaxation.

            The second value is the current task, where 'continuedir' is the path to a
            vasp job directory that is not yet completed, "relax" indicates another
            volume relaxation job is required, and "constant" that a constant volume run is required.
        """

        # check if all complete
        if io.job_complete(self.finaldir):
            return ("complete", None)

        # check status of relaxation runs
        self.update_rundir()

        # if not yet started
        if len(self.rundir) == 0:
            return ("incomplete", "setup")

        # if the latest run is complete:
        if io.job_complete(self.rundir[-1]):

            # if it is a final constant volume run
            if io.get_incar_tag("SYSTEM", self.rundir[-1]) != None:
                if io.get_incar_tag("SYSTEM",
                                    self.rundir[-1]).split()[-1].strip().lower(
                                    ) == "final":
                    # if io.get_incar_tag("ISIF", self.rundir[-1]) == 2 and \
                    #    io.get_incar_tag("NSW", self.rundir[-1]) == 0 and \
                    #    io.get_incar_tag("ISMEAR", self.rundir[-1]) == -5:
                    return ("complete", None)

            # elif constant volume run (but not the final one)
            if io.get_incar_tag("ISIF", self.rundir[-1]) in [0, 1, 2]:
                if io.get_incar_tag("NSW", self.rundir[-1]) == len(
                        io.Oszicar(os.path.join(self.rundir[-1],
                                                "OSZICAR")).E):
                    return ("incomplete", "relax"
                            )  # static run hit NSW limit and so isn't "done"
                else:
                    return ("incomplete", "constant")

            # elif convergence criteria met
            if self.converged():
                return ("incomplete", "constant")

            # elif not converging, return 'not_converging' error
            if self.not_converging():
                return ("not_converging", None)

            # else continue relaxing
            else:
                return ("incomplete", "relax")

        # elif not converging, return 'not_converging' error
        elif self.not_converging():
            return ("not_converging", None)

        # else if the latest run is not complete, continue running it
        return ("incomplete", self.rundir[-1])



from setuptools import setup, find_packages
import glob
import os
from casm import __version__


# get console_scripts
def script_str(file):
    name = os.path.splitext(os.path.split(file)[1])[0]
    if name in ['casm_calc', 'casm_learn']:
        return name.replace('_', '-') + '=casm.scripts.' + name + ':main'
    else:
        return name.replace('_', '.') + '=casm.scripts.' + name + ':main'


console_scripts = [
    script_str(x) for x in glob.glob('casm/scripts/*.py')
    if x != 'casm/scripts/__init__.py'
]
print(console_scripts)

with open(os.path.join('..', 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='casm-python',
    version=__version__,
    url='https://github.com/prisms-center/CASMpython',
    description='CASM Python interface, tools, and wrappers.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='CASM developers',
    author_email='casm-developers@lists.engr.ucsb.edu',
    license='LGPL2.1+',
    packages=find_packages(),
    entry_points={'console_scripts': console_scripts},
    install_requires=[
        'deap', 'mock', 'pandas', 'prisms-jobs', 'scikit-learn',
        'scipy', 'sh'
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering'
    ],
    data_files=[('', ['LICENSE'])])



import json
import os


class AttributeInfoError(Exception):
    """Exception handling"""
    def __init__(self, message):
        self.message = message

    def __str__(self):
        """ Writes out the error message

        Returns
        -------
        string

        """
        return self.message


class AttributeInfo:
    """Contains the atom and molecule attributes' information
       self.atom_type: List[str] - Contains the atom types
       self.atom_dofs: Dict{Dict{List[float or int]}} - Contains dofs for all the atoms in the structure.json
       self.mol_type: List[str] - Contains the molecule types
       self.mol_dofs: Dict{Dict{List[float or int]}} - Contains dofs for the molecules
       Look the example below for clarity regarding object properties

       For example consider a Cmagspin dof for NaFeO2 system
       self.atom_type: [Na, Fe, O, O]
       self.atom_dofs: {'Cmagspin':{'value': [[0],[-1],[0],[0]]}}
       self.mol_type: [Na, Fe-, O, O]
       self.mol_dofs: {'Cmagspin': {'value': [[0],[0],[0],[0]]}}"""
    def __init__(self, filename):
        """Constructs the AttributeInfo object by reading the structure.json file

        Parameters
        ----------
        filename : string (structure.json file)

        """
        self.read(filename)

    def read(self, filename):
        """Reads the attribute information from a CASM structure.json file

        Will set attributes to None, with message "No dof information found in <filename>", if the file cannot be read as a CASM structure.json file.

        Parameters
        ----------
        filename : string (structure.json file)

        """
        try:
            with open(filename, 'r') as f:
                config_data = json.load(f)

            self.atom_dofs = config_data["atom_dofs"]
            self.atom_type = config_data["atom_type"]
            self.mol_dofs = config_data["mol_dofs"]
            self.mol_type = config_data["mol_type"]

        except:
            self.atom_dofs = None
            self.atom_type = None
            self.mol_dofs = None
            self.mol_type = None
            print("No dof information found in " + filename)



import json
from casm.api import casm_capture


def _check_info_output(stdout, returncode, method):
    if returncode:
        print(stdout)
        raise Exception("Error getting " + method + " info")
    try:
        return json.loads(stdout)
    except:
        print(stdout)
        raise Exception("Error parsing " + method + " info results")


def prim_info_desc():
    """
    Return CASM `info --desc PrimInfo` output.
    """
    stdout, returncode = casm_capture('info --desc PrimInfo',
                                      combine_output=True)
    return stdout


def get_prim_info(prim=None, root=None, properties=[]):
    """
    Get properties of a primitive crystal structure.

    Arguments
    ---------

      prim: dict of BasicStructure (optional, default=None)
        Dict containing the definition of a "prim", in the `CASM BasicStructure JSON format`_. May be None if `root` provided.

      root: str (optional, default=os.getcwd())
        Use the prim of a CASM project located at `root`, if it exists.

      properties: list of str (default=[])
        A list of properties to return


    Returns
    -------
      output: Dict containing results.

    .. CASM BasicStructure JSON format_: https://prisms-center.github.io/CASMcode_docs/pages/formats/casm/crystallography/BasicStructure.html
    """
    input = {}
    if prim is not None:
        input["prim"] = prim
    input["properties"] = properties
    args = "info -m PrimInfo -i '" + json.dumps(input) + "'"
    stdout, returncode = casm_capture(args, root=root, combine_output=True)
    return _check_info_output(stdout, returncode, "prim")


def supercell_info_desc():
    """
    Return CASM `info --desc SupercellInfo` output.
    """
    stdout, returncode = casm_capture('info --desc SupercellInfo',
                                      combine_output=True)
    return stdout


def get_supercell_info(prim=None, root=None, properties=[], **kwargs):
    """
    Get properties of a supercell.

    Arguments
    ---------

      prim: dict of BasicStructure (optional, default=None)
        Dict containing the definition of a "prim", in the `CASM BasicStructure JSON format`_. May be None if `root` provided.

      root: str (optional, default=os.getcwd())
        Use the prim of a CASM project located at `root`.

      properties: list of str (default=[])
        A list of properties to return

      **kwargs:  key-value pairs
        Additional key-value pairs to be added to the input. See
        `supercell_info_desc` for options.

    Returns
    -------
      output: Dict containing results.

    .. CASM BasicStructure JSON format_: https://prisms-center.github.io/CASMcode_docs/pages/formats/casm/crystallography/BasicStructure.html
    """
    input = {}
    if prim is not None:
        input["prim"] = prim
    input["properties"] = properties
    input.update(kwargs)
    args = "info -m SupercellInfo -i '" + json.dumps(input) + "'"
    stdout, returncode = casm_capture(args, root=root, combine_output=True)
    return _check_info_output(stdout, returncode, "supercell")


def neighbor_list_info_desc():
    """
    Return CASM `info --desc NeighborListInfo` output.
    """
    stdout, returncode = casm_capture('info --desc NeighborListInfo',
                                      combine_output=True)
    return stdout


def get_neighbor_list_info(prim=None, root=None, properties=[], **kwargs):
    """
    Get neighbor list information for a supercell

    Arguments
    ---------

      prim: dict of BasicStructure (optional, default=None)
        Dict containing the definition of a "prim", in the `CASM BasicStructure JSON format`_. May be None if `root` provided.

      root: str (optional, default=os.getcwd())
        Use the prim of a CASM project located at `root`.

      properties: list of str (default=[])
        A list of properties to return

      **kwargs:  key-value pairs
        Additional key-value pairs to be added to the input. See
        `neighbor_list_info_desc` for options.

    Returns
    -------
      output: Dict containing results.

    .. CASM BasicStructure JSON format_: https://prisms-center.github.io/CASMcode_docs/pages/formats/casm/crystallography/BasicStructure.html
    """
    input = {}
    if prim is not None:
        input["prim"] = prim
    input["properties"] = properties
    input.update(kwargs)
    args = "info -m NeighborListInfo -i '" + json.dumps(input) + "'"
    stdout, returncode = casm_capture(args, root=root, combine_output=True)
    return _check_info_output(stdout, returncode, "neighbor_list")



import numpy as np


def canonical_unroll_index_list_recursive(n, origin, current_list):
    """Recursive helper function for canonical_unroll_index_list"""
    o_i, o_j = origin
    for k_sw in range(n):
        current_list.append((o_i + k_sw, o_j + k_sw))
    for k_n in range(n - 1):
        current_list.append((o_i + n - 2 - k_n, o_j + n - 1))
    for k_w in range(n - 2):
        current_list.append((o_i, o_j + n - 2 - k_w))
    if n >= 3:
        return canonical_unroll_index_list_recursive(n - 3, (o_i + 1, o_j + 2),
                                                     current_list)
    else:
        return current_list


def canonical_unroll_index_list(n):
    """Returns an ordered list of indices defining the canonical unrolling of an n by n matrix"""
    return canonical_unroll_index_list_recursive(n, (0, 0), [])


def vectorized_index(i, j, n):
    """Returns the index corresponding to entry i,j in a column-wise vectorization of an n by n matrix"""
    return n * i + j


def reduction_matrix(n):
    """Returns the m by n^2 reduction_matrix matrix that transforms the dimension n^2
        vectorization V of a symmetric n by n matrix U into a vector T of
        dimension m corresponding to the independent elements of U,
        unrolled canonically, with the off-diagonal elements being multiplied by sqrt(2)
    """
    off_diagonal = 1 / np.sqrt(2)
    index_list = canonical_unroll_index_list(n)
    m = len(index_list)
    reduction_matrix = np.zeros((m, n**2))
    for k, (i, j) in enumerate(index_list):
        if i == j:
            reduction_matrix[k, vectorized_index(i, j, n)] = 1
        else:
            reduction_matrix[k, vectorized_index(i, j, n)] = off_diagonal
            reduction_matrix[k, vectorized_index(j, i, n)] = off_diagonal
    return reduction_matrix


def unroll_symmetric_matrix(matrix):
    """Returns the vector that is the unrolled version of the symmetrix matrix given"""
    matrix = np.array(matrix)
    m, n = matrix.shape
    assert (m == n)
    vectorized = matrix.flatten(order='F')
    reduction = reduction_matrix(n)
    return np.matmul(reduction, vectorized)


def reroll_symmetric_matrix(unrolled):
    """Returns the re-rolled symmetric matrix corresponding to the given vector"""
    unrolled = np.array(unrolled)
    n = int(np.sqrt(2 * len(unrolled) + 0.25) -
            0.5)  # inverse of triangular number
    reduction = reduction_matrix(n)
    vectorized = np.matmul(reduction.transpose(), unrolled)
    return np.reshape(vectorized, (n, n), order='F')


def is_zero(array, tol=1e-08):
    """Returns True if the given array is equal to zero, within the given tolerance"""
    return np.allclose(array, np.zeros(array.shape), atol=tol)



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import os
import re
import shutil
import six
import sys
from casm.vasp.io import oszicar, outcar, species, poscar
from casm.vasp.io import kpoints as kp
from casm.vasp.io import incar as inc
from casm.project import attribute_info

VASP_INPUT_FILE_LIST = [
    "INCAR", "STOPCAR", "POTCAR", "KPOINTS", "POSCAR", "EXHCAR", "CHGCAR",
    "WAVECAR", "TMPCAR"
]

DEFAULT_VASP_MOVE_LIST = ["POTCAR"]

DEFAULT_VASP_COPY_LIST = ["INCAR", "KPOINTS"]

DEFAULT_VASP_REMOVE_LIST = [
    "IBZKPT", "CHG", "CHGCAR", "WAVECAR", "TMPCAR", "EIGENVAL", "DOSCAR",
    "PROCAR", "PCDAT", "XDATCAR", "LOCPOT", "ELFCAR", "PROOUT"
]


class VaspIOError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


def job_complete(jobdir=None):
    """Return True if vasp job at path 'jobdir' is complete"""
    if jobdir is None:
        jobdir = os.getcwd()
    outcarfile = os.path.join(jobdir, "OUTCAR")
    if (not os.path.isfile(outcarfile)) and (
            not os.path.isfile(outcarfile + ".gz")):
        return False
    if outcar.Outcar(outcarfile).complete:
        return True
    return False


def get_incar_tag(key, jobdir=None):
    """Opens INCAR in 'jobdir' and returns 'key' value."""
    if jobdir is None:
        jobdir = os.getcwd()
    tincar = inc.Incar(os.path.join(jobdir, "INCAR"))
    for k in tincar.tags:
        if key.lower() == k.lower():
            return tincar.tags[k]
    return None


def set_incar_tag(tag_dict, jobdir=None, name=None):
    """Opens INCAR in 'jobdir', sets 'key' value, and writes INCAR
        If 'val' is None, the tag is removed from the INCAR.
    """
    if name is None:
        name = "INCAR"
    if jobdir is None:
        jobdir = os.getcwd()
    incarfile = os.path.join(jobdir, name)
    tincar = inc.Incar(incarfile)

    for key, val in six.iteritems(tag_dict):
        for k in tincar.tags:
            if key.lower() == k.lower():
                if (val is None) or (str(val).strip() == ""):
                    del tincar.tags[k]
                else:
                    tincar.tags[k] = val
                break

        if val != None and str(val).strip() != "":
            tincar.tags[key] = val

    tincar.write(incarfile)


def ionic_steps(jobdir=None):
    """Find the number of ionic steps completed in 'jobdir'"""
    try:
        toszicar = oszicar.Oszicar(os.path.join(jobdir, "OSZICAR"))
        return len(toszicar.E)
    except:
        raise VaspIOError("Could not read number of ionic steps from " +
                          os.path.join(jobdir, "OSZICAR"))


def write_potcar(filename, poscar, species, sort=True):
    """ Write an appropriate POTCAR """
    if sort == False:
        with open(filename, 'w') as file:
            for name in poscar.type_atoms:
                with open(os.path.join(species[name].potcardir,
                                       'POTCAR')) as potcar:
                    file.write(potcar.read())
    else:
        # dict: key = alias, value = list of Sites
        pos = poscar.basis_dict()

        with open(filename, 'w') as file:
            # for each alias
            for alias in sorted(pos.keys()):
                # find matching IndividualSpecies with write_potcar == True
                for name in species:
                    if species[name].alias == alias and species[
                            name].write_potcar:
                        # add to POTCAR file
                        with open(
                                os.path.join(species[name].potcardir,
                                             'POTCAR')) as potcar:
                            file.write(potcar.read())
                        break


def write_stopcar(mode='e', jobdir=None):
    """ Write STOPCAR file with two modes:
        mode = 'e' for 'VASP stops at the next electronic step'
        mode = 'i' for 'VASP stops at the next ionic step' """
    if jobdir is None:
        jobdir = os.getcwd()
    if mode.lower()[0] == 'e':
        stop_string = "LABORT = .TRUE."
    elif mode.lower()[0] == 'i':
        stop_string = "LSTOP = .TRUE."
    else:
        raise VaspIOError("Invalid STOPCAR mode specified: " + str(mode))

    filename = os.path.join(jobdir, 'STOPCAR')

    try:
        stopcar_write = open(filename, 'w')
    except IOError as e:
        raise e

    stopcar_write.write(stop_string)
    stopcar_write.close()


def write_vasp_input(dirpath,
                     incarfile,
                     ref_kpointsfile,
                     ref_structurefile,
                     structurefile,
                     speciesfile,
                     sort=True,
                     extra_input_files=[],
                     strict_kpoints=False):
    """ Write VASP input files in directory 'dirpath'

    Parameters
    ----------

    ref_structurefile: str
        Path to a CASM structure.json file or VASP POS/POSCAR file representing a reference structure used for scaling incar and k-point parameters.

    structurefile: str
        Path to a CASM structure.json file or VASP POS/POSCAR file representing the structure to be calculated.
    """
    print("Setting up VASP input files:", dirpath)

    # read reference structure and kpoints
    print("  Reading reference KPOINTS:", ref_kpointsfile)
    ref_kpoints = kp.Kpoints(ref_kpointsfile)
    if ref_structurefile != None:
        print("  Reading reference POSCAR:", ref_structurefile)
        ref_structure = poscar.Poscar(ref_structurefile)
    else:
        ref_structure = None

    # read species, prim structure.json/POS, and to-be-calculated
    # structure.json/POS, and use to construct incar and kpoints for
    # the to-be-calculated structure
    print("  Reading SPECIES:", speciesfile)
    species_settings = species.species_settings(speciesfile)
    if structurefile != None:
        print("  Reading structure:", structurefile)
        structure = poscar.Poscar(structurefile, species_settings)
    else:
        structure = None

    ## Reading DOF information present in the structure.json file
    ## Examples include: Cmagspin, NCmagspin, SOmagspin, etc
    ## Also contains information about atom types, and mol types
    print(" Reading DOF information from structure: ", structurefile)
    dof_info = attribute_info.AttributeInfo(structurefile)

    print("  Reading INCAR:", incarfile)
    incar = inc.Incar(incarfile, species_settings, structure, sort, dof_info)

    print("  Generating KPOINTS")
    if strict_kpoints:
        kpoints = ref_kpoints
    else:
        kpoints = ref_kpoints.super_kpoints(ref_structure, structure)

    # write main input files
    if structurefile != None:
        print("  Writing POSCAR:", os.path.join(dirpath, 'POSCAR'))
        structure.write(os.path.join(dirpath, 'POSCAR'), sort)
    print("  Writing INCAR:", os.path.join(dirpath, 'INCAR'))
    incar.write(os.path.join(dirpath, 'INCAR'))
    print("  Writing KPOINTS:", os.path.join(dirpath, 'KPOINTS'))
    kpoints.write(os.path.join(dirpath, 'KPOINTS'))
    if structurefile != None:
        print("  Writing POTCAR:", os.path.join(dirpath, 'POTCAR'))
        write_potcar(os.path.join(dirpath, 'POTCAR'), structure,
                     species_settings, sort)
    # copy extra input files
    if len(extra_input_files):
        print("  Copying extra input files", end=' ')
    for s in extra_input_files:
        print("    ", s)
        shutil.copy(s, dirpath)

    print("  DONE\n")
    sys.stdout.flush()



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import numpy as np
import random


def eci(individual, coef):
    """ 
  Return ECI as list of tuple containing (index, coef).
  
  Arguments
  ---------
    
    individual: List[bool] of length n_features
      This is a boolean list of shape [n_features], in which an element is True 
      iff its corresponding feature is selected for retention.
    
    coef: array-like of shape (n_selected, 1)
      The non-zero coefficients.
  
  
  Returns
  -------
    
    eci: List[(int, float)]
      A list of tuple containing the basis function index and coefficient value
      for basis functions with non-zero coefficients: [(index, coef), ...]
  
  """
    return list(zip(indices(individual), list(coef)))


def indices(individual):
    """ 
  Convert list of bool to list of indices of True values.
  
  Arguments
  ---------
    
    individual: List[bool] of length n_features
      This is a boolean list of shape [n_features], in which an element is True 
      iff its corresponding feature is selected for retention.
    
  
  Returns
  -------
    
    indices: List[int]
      List of indices of True values.
  
  """
    return [i for i in range(len(individual)) if individual[i]]


def wHullDist(hull_dist, A=1.0, B=1.0, kT=1.0, **kwargs):
    """
  Returns A*np.exp(-hull_dist/kT) + B
  """
    return A * np.exp(-hull_dist / kT) + B


def wEmin(y, A=1.0, B=1.0, kT=1.0):
    """
  Returns A*np.exp(-(y - emin)/kT) + B, where emin = np.min(y)
  """
    emin = np.min(y)
    return A * np.exp(-(y - emin) / kT) + B


def wEref(y, A=1.0, B=1.0, kT=1.0, E0=0.0):
    """
  Returns A*np.exp(-(y - E0)/kT) + B, where y >= E0; 1.0 otherwise
  """
    w = A * np.exp(-(y - E0) / kT) + B
    w[list(np.where(y < E0)[0])] = 1.0
    return w


def set_sample_weight(sample_weight, y=None, X=None):
    """ 
  Calculate weighted data and weighted target values.
  
  Ordinary least squares minimizes
    (y-X*b).transpose() * (y-X*b)
  
  where 'X' is the correlation matrix of shape (Nvalue, Nbfunc), and 'y' 
  is a vector of Nvalue calculated properties, and 'b' are the fitting 
  coefficients (ECI).

  Weighted least squares minimizes
    (y-X*b).transpose() * W * (y-X*b)
  
  Using the SVD, and given that W is Hermitian:
    U * S * U.transpose() == W
  
  Define L such that:
    L.transpose() = U * sqrt(S)
  
  Then we can write the weighted least squares problem using:
    (y-X*b).transpose() * L.transpose() * L * (y-X*b)
  
  Or:
    (L*y-L*X*b).transpose() * (L*y-L*X*b)
    
  So, if weights are included, then the linear model is changed from
    X*b = y  ->  L*X*b = L*y
    
    
  Arguments
  ---------
    
    sample_weight: None, 1d array-like of shape (n_samples,1), or 2d array-like of shape (n_samples, n_samples)
      Sample weights.
      
      if sample_weight is None: (unweighted)
        W = np.matlib.eye(N) 
      if sample_weight is 1-dimensional:
        W = np.diag(sample_weight)*Nvalue/np.sum(sample_weight) 
      if sample_weight is 2-dimensional (must be Hermitian, positive-definite):
        W = sample_weight*Nvalue/np.sum(sample_weight)
    
    X: array-like of shape (n_samples, n_features), optional
      The training input samples (correlations).
    
    y: array-like of shape: (n_samples, 1), optional
      The target values (property values).
    
  Returns
  -------
    
    (weighted_y, weighted_X, W, L)
    
    weighted_y: None, or array-like of shape: (n_samples, 1)
      If y given as input, the weighted target values, weighted_y = L*y. 
    
    weighted_X: None, or array-like of shape: (n_samples, n_features)
      If X given as input, the weighted training input data, weighted_X = L*x.
    
    W: array-like of shape: (n_samples, n_samples)
      Contains sample weights. 
    
    L: array-like of shape: (n_samples, n_samples)
      Used to generate weighted_X and weighted_y, W = L * L.transpose(). 
    
      
  Notes
  -----
    
  Returns (weighted_y, weighted_X, W, L) 
  """
    # check sample_weight and convert to square matrix
    W = None
    L = None
    weighted_y = None
    weighted_X = None

    if sample_weight is None:
        W = np.identity(y.shape[0])
    elif len(sample_weight.shape) == 1:
        n_samples = len(sample_weight)
        W = np.diag(sample_weight) * n_samples / np.sum(sample_weight)
    elif len(sample_weight.shape) == 2:
        n_samples = len(sample_weight)
        W = sample_weight * n_samples / np.sum(sample_weight)
    else:
        raise Exception(
            "Error in set_sample_weight: sample_weight dimension > 2")

    # weighted data
    U, S, V = np.linalg.svd(W)
    L = U.dot(np.diag(np.sqrt(S))).transpose()

    if X is not None:
        weighted_X = np.dot(L, X)

    if y is not None:
        weighted_y = np.dot(L, y)

    return (weighted_y, weighted_X, W, L)



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import xml.etree.cElementTree as etree
import os, gzip
import numpy as np


class VasprunError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


class Vasprun:
    """ An object containing values read from vasprun.xml

        The Vasprun class contains:
            self.total_energy: final total energy ('e_wo_entrp')
            self.forces: final forces on atoms (2d list of double)
            self.atom_type: list of atom types (list of str)
            self.atoms_per_type: list of number of atoms for each atom type (list of int)
            self.lattice:  final lattice (2d list of double)
            self.rec_lat: final reciprocal lattice (2d list of double)
            self.basis: final basis (2d list of double)
            self.coord_mode: coordinate mode for basis (always 'direct')
            self.is_complete: did the VASP calculation run to completion? (bool)
            self.dos: Basic DOS output (3d list of double or bool)
            self.dos_efermi: Fermi level for DOS use (double)
            self.dos_lm: l or lm-projected DOS output (4d list of double or bool)
            self.eigenvalues: eigenvalues and energies for doing band structure plots (list of 2d list of double)
            self.all_e_0: energy (e_0_energy) for each electronic step of each ionic step
            self.nelm: NELM (max. number of electronic steps)
    """
    def __init__(self, filename, DOS=False, Band=False):
        """ Create a Vasprun object from a vasprun.xml file with name 'filename' """

        #### Filename #########################
        self.filename = filename

        #### set default values ###############

        # final energy ('e_wo_entrp')
        self.total_energy = None

        # final forces on atoms (2d list)
        self.forces = []

        # list of atom types
        self.atom_type = []

        # list of number of atoms for each atom type
        self.atoms_per_type = []

        # final lattice (2d list)
        self.lattice = None

        # final rec lattice (2d list)
        self.rec_lat = None

        # final basis (2d list)
        self.basis = None

        # coordinate mode for basis
        self.coord_mode = 'direct'

        # did the VASP calculation run to completion?
        self.is_complete = False

        # Is DOS present?
        self.DOS = DOS  # Whether to even read DOS; grabbed from init args
        self.dos = None

        # What about band structure?
        self.Band = Band
        if self.Band:
            self.DOS = True  # Need to read DOS to do some band things

        # What is the fermi level?
        self.efermi = None

        # is l or lm-projected DOS present?
        self.dos_lm = None

        # Energy (e_0_energy) for each electronic step for each ionic step
        self.all_e_0 = []

        ##### read from vasprun.xml file ########
        self.iter_read()

    def is_complete(self):
        """ Return True if VASP calculation ran to completion """
        return self.is_complete

    def iter_read(self):
        """ Create a Vasprun object from a vasprun.xml file with name 'filename' """
        if os.path.isfile(self.filename):
            if self.filename.split(".")[-1].lower() == "gz":
                f = gzip.open(self.filename)
            else:
                f = open(self.filename)
        elif os.path.isfile(self.filename + ".gz"):
            f = gzip.open(self.filename + ".gz")
        else:
            raise VasprunError("file not found: " + self.filename)

        for event, elem in etree.iterparse(f):
            if elem.tag == 'parameters':
                self.nelm = int(elem.find(".//i[@name='NELM']").text)

            if elem.tag == 'calculation':

                #finding the energy of the relaxation
                self.total_energy = None
                energy = elem.findall('energy')
                for child in energy[-1]:
                    if child.attrib['name'] == 'e_wo_entrp':
                        self.total_energy = float(child.text)
                        break
                    child.clear()

                #finding the forces on the atoms
                varray = elem.findall('varray')
                forces = None
                for v in varray:
                    if v.attrib['name'] == 'forces':
                        forces = v
                self.forces = []
                for child in forces:
                    self.forces.append(
                        [float(i) for i in child.text.strip().split()])
                    child.clear()
                v.clear()

                # find energy (e_0_energy) for each electronic step in this ionic step
                self.all_e_0.append([])
                for i in elem.findall("./scstep/energy/i[@name='e_0_energy']"):
                    self.all_e_0[-1].append(float(i.text))

                if self.DOS:

                    # gather the DOS, if calculated and warranted
                    self.dos = None
                    self.dos_lm = None
                    self.efermi = None

                    is_dos = elem.find('dos')
                    if is_dos != None:
                        for child in is_dos:
                            if child.attrib['name'] == 'efermi':
                                self.efermi = float(child.text)
                                child.clear()
                                break
                            else:
                                child.clear()

                        total = is_dos.find('total').find('array')

                        spins = len(total.find('set').findall('set'))
                        points = len(
                            total.find('set').findall('set')[0].findall('r'))
                        fields = len(total.findall('field'))

                        total_array = np.zeros([spins, points, fields])

                        for s in range(spins):
                            my_set = total.find('set').findall(
                                'set')[s].findall('r')
                            for r in range(points):
                                total_array[s,
                                            r] = map(float,
                                                     my_set[r].text.split())

                        self.dos = total_array

                        partial = is_dos.find('partial')
                        if partial != None:
                            partial = partial.find('array')
                        else:
                            partial = None

                        if partial:
                            ions = len(partial.find('set').findall('set'))
                            spins = len(
                                partial.find('set').findall('set')[0].findall(
                                    'set'))
                            points = len(
                                partial.find('set').findall('set')[0].findall(
                                    'set')[0].findall('r'))
                            fields = len(partial.findall('field'))

                            partial_array = np.zeros(
                                [ions, spins, points, fields])

                            for i in range(ions):
                                i_set = partial.find('set').findall(
                                    'set')[i].findall('set')
                                for s in range(spins):
                                    my_set = i_set[s].findall('r')
                                    for r in range(points):
                                        partial_array[i, s, r] = map(
                                            float, my_set[r].text.split())

                            self.dos_lm = partial_array

                if self.Band:
                    eig_parse_vals = elem.find('eigenvalues').find(
                        'array').find('set')
                    num_kpoints = len(
                        eig_parse_vals.find('set').findall('set'))
                    num_bands = len(
                        eig_parse_vals.find('set').find('set').findall('r'))
                    self.eigenvalues = [
                        np.empty([num_kpoints, num_bands]),
                        np.empty([num_kpoints, num_bands])
                    ]
                    for spin_ctr, spin in enumerate(eig_parse_vals):
                        for kpoint_ctr, kpoint in enumerate(spin):
                            for band_ctr, band in enumerate(kpoint):
                                self.eigenvalues[spin_ctr][
                                    kpoint_ctr,
                                    band_ctr] = band.text.strip().split()[0]
                    eig_parse_vals.clear()

            elif elem.tag == 'kpoints':
                if self.Band:
                    self.num_divisions = elem.find('generation').find('i')
                    if self.num_divisions != None:
                        self.num_divisions = int(
                            self.num_divisions.text.strip())
                        kp_parse_vals = elem.find('varray') if elem.find(
                            'varray').attrib['name'] == 'kpointlist' else None
                        if kp_parse_vals != None:
                            num_paths = len(kp_parse_vals.findall(
                                'v')) / self.num_divisions
                            kpoint_list = np.zeros(
                                [self.num_divisions * num_paths, 3])
                            for i, row in enumerate(kp_parse_vals):
                                kpoint_list[i, :] = [
                                    float(val)
                                    for val in row.text.strip().split()
                                ]
                            self.kpoint_divisions = np.empty(
                                [2 * num_paths, 3])
                            for i in range(num_paths):
                                self.kpoint_divisions[2 * i] = kpoint_list[
                                    i * self.num_divisions, :]
                                self.kpoint_divisions[2 * i + 1] = kpoint_list[
                                    (i + 1) * self.num_divisions - 1, :]
                            kp_parse_vals.clear()
                    else:
                        self.kpoint_divisions = False

            elif elem.tag == 'atominfo':

                #finding the atom_type and atoms_per_type
                temp_node = elem.findall('array')
                for child in temp_node:
                    if child.attrib['name'] == 'atomtypes':
                        child_node = child
                        break
                t_atom_info = child_node.findall('set')[0].findall('rc')
                self.atom_type = []
                self.atoms_per_type = []
                for child in t_atom_info:
                    self.atom_type.append(child[1].text)
                    self.atoms_per_type.append(int(child[0].text))

            elif elem.tag == 'structure':
                self.is_complete = False
                if 'name' in elem.attrib:
                    if elem.attrib['name'] == 'finalpos':
                        #finding the final structure
                        finalpos = elem
                        self.is_complete = True

                        # collect the final lattice
                        self.lattice = []
                        for i in range(3):
                            self.lattice.append([
                                float(x) for x in finalpos[0][0]
                                [i].text.strip().split()
                            ])

                        # Collect the reciprocal lattice
                        self.rec_lat = []
                        for i in range(3):
                            self.rec_lat.append([
                                float(x) for x in finalpos[0][2]
                                [i].text.strip().split()
                            ])

                        # collect the final basis
                        self.basis = []
                        for i, basis_atom in enumerate(finalpos[1]):
                            self.basis.append([
                                float(x)
                                for x in basis_atom.text.strip().split()
                            ])
                        finalpos.clear()

            else:
                pass
        elem.clear()
        f.close()



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import numpy as np
import sklearn


class LinearRegressionForLOOCV(sklearn.base.BaseEstimator):
    """ 
  LinearRegression estimator that fits by constructing hat matrix for faster
  LOOCV score calculations.
  
  Attributes
  ----------
    
    pinv: boolean, optional, default=True
      If true, use the psuedo-inverse in solving the least squares problem.
    
    coef_: array-like of shape (n_samples, 1)
      Estimated coefficients for the linear regression problem.
        
    H_: array-like of shape (n_samples, n_features)
      H = X*(X.transpose()*X).inverse()*X.transpose()
  """
    def __init__(self, pinv=True, **kwargs):
        """
    Calculates linear least squares solution for X*b = y.
    
    Arguments
    ---------
    
      pinv: boolean, optional, default=True
        If true, use the psuedo-inverse in solving the least squares problem.
      
      kwargs: keyword args
        Other keyword args are ignored
    """
        self.pinv = pinv

    def fit(self, X, y):
        """
    Calculates linear least squares solution for X*b = y.
    
    
    Arguments
    ---------
    
      X: array-like of shape (n_samples, n_features)
        The input data
      
      y: array-like of shape (n_samples, 1)
        The values
      
    
    Notes
    -----
    
      Derivation:
        X.t*X*b = X.t*y, orthogonal projection
        b = (X.t*X).inv * X.t * y, solves the least squares problem
        b = S * y,
      where
        S = ( X.t * X ).inv * X.t
      
      and 
        y_pred = X * b = H * y,
      where
        H = X * S
      
      Calculates attributes: coef_, H_
    """
        if self.pinv == False:
            S = np.linalg.inv(X.transpose().dot(X)).dot(X.transpose())
        else:
            S = np.linalg.pinv(X.transpose().dot(X)).dot(X.transpose())

        # coef
        self.coef_ = np.dot(S, y)

        # stores results for LOOCV formula
        self.H_ = np.dot(X, S)

    def predict(self, X):
        """
    Predict using the linear model.
    
    
    Arguments
    ---------
    
      X: array-like of shape (n_samples, n_features)
        The input data
    
    
    Returns
    -----
    
      y_pred: array-like of shape (n_samples, 1)
        The predicted values
    """
        return np.dot(X, self.coef_)

    def score(self, X, y):
        """
    Return the LOOCV score using the linear model.
    
    
    Arguments
    ---------
    
      X: array-like of shape (n_samples, n_features)
        The input data. (This parameter is ignored)
      
      y: array-like of shape (n_samples, 1)
        The values
      
    
    Returns
    -----
    
      loocv: float
        The LOOCV score, np.mean(((y - y_pred)/(1.0 - np.diag(self.H_)))**2)
    
    Notes
    -----
      Must already be fit. 'X' parameter is ignored.
      
    """
        y_pred = np.dot(self.H_, y)
        return np.mean(((y - y_pred) / (1.0 - np.diag(self.H_)))**2)



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *
from casm import vasp

from casm import vasp
import os, shutil, six, re, subprocess, json
import warnings
import casm.vasp.io


class VaspWrapperError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


def read_settings(filename):
    """Returns a JSON object reading JSON files containing settings for VASP PBS jobs.

   Returns:
        settings = a JSON object containing the settings file contents
                     This can be accessed like a dict: settings["account"], etc.
                     ** All values are expected to be 'str' type. **

    The required keys are:
        "queue": queue to submit job in
        "ppn": processors (cores) per node to request
        "walltime": walltime to request (ex. "48:00:00")

    The optional keys are:
        "atom_per_proc": max number of atoms per processor (core)
        "account": account to submit job under (default None)
        "pmem": string for requested memory (default None)
        "priority": requested job priority (default "0")
        "constraint": constraint. ex: ``"haswell"`` (default None)
        "message": when to send messages about jobs (ex. "abe", default "a")
        "email": where to send messages (ex. "me@fake.com", default None)
        "qos": quality of service, 'qos' option (ex. "fluxoe")
        "npar": vasp incar setting (default None)
        "ncore": vasp incar setting (default None)
        "kpar": vasp incar setting (default None)
        "vasp_cmd": vasp execution command (default is "vasp" (ncpus=1) or "mpirun -np {NCPUS} vasp" (ncpus!=1))
        "ncpus": number of cpus (cores) to run on (default $PBS_NP)
        "run_limit": number of vasp runs until "not_converging" (default 10)
        "nrg_convergence": converged if last two runs complete and differ in energy by less than this amount (default None)
        "move": files to move at the end of a run (ex. ["POTCAR", "WAVECAR"], default ["POTCAR"])
        "copy": files to copy from run to run (ex. ["INCAR", "KPOINTS"], default ["INCAR, KPOINTS"])
        "remove": files to remove at the end of a run (ex. ["IBZKPT", "CHGCAR"], default ["IBKZPT", "CHG", "CHGCAR", "WAVECAR", "TMPCAR", "EIGENVAL", "DOSCAR", "PROCAR", "PCDAT", "XDATCAR", "LOCPOT", "ELFCAR", "PROOUT"]
        "compress": files to compress at the end of a run (ex. ["OUTCAR", "vasprun.xml"], default [])
        "backup": files to compress to backups at the end of a run, used in conjunction with move (ex. ["WAVECAR"])
        "extra_input_files": extra input files to be copied from the settings directory, e.g., a vdW kernel file.
        "initial" : location of INCAR with tags for the initial run, if desired (e.g. to generate a PBE WAVECAR for use with M06-L)
        "final" : location of INCAR with tags for the final run, if desired (e.g. "ISMEAR = -5", etc). Otherwise, the settings enforced are ("ISMEAR = -5", "NSW = 0", "IBRION = -1", "ISIF = 2")
        "err_types" : list of errors to check for. Allowed entries are "IbzkptError" and "SubSpaceMatrixError". Default: ["SubSpaceMatrixError"]
        "preamble" : a text file containing anything that MUST be run before python is invoked (e.g. module.txt which contains "module load python", or "source foo")
        "prerun" : bash commands to run before vasp.Relax.run (default None)
        "postrun" : bash commands to run after vasp.Relax.run completes (default None)
        "prop": USED IN vasp.converge ONLY. Property to converge with respect to (current options are "KPOINT" and "ENCUT")
        "prop_start": USED IN vasp.converge ONLY. Starting value of "prop", e.g. 450 (for ENCUT) or 5 (for KPOINTS) or [4 4 4] (for KPOINTS)
        "prop_stop": USED IN vasp.converge ONLY. Ending value of "prop", e.g. 550 (for ENCUT) or 20 (for KPOINTS).
        "prop_step": USED IN vasp.converge ONLY. Delta value of "prop", e.g. 10 (for ENCUT) or 2 (for KPOINTS) or [1 1 2] (for KPOINTS)
        "tol" : USED IN vasp.converge ONLY. Tolerance type for convergence, e.g. relaxed_energy. Optional
        "tol_amount" : USED IN vasp.converge ONLY. Tolerance criteria convergence, e.g. 0.001. If the abs difference between two runs in their "tol" is smaller than "tol_amount", the "prop" is considered converged.
        "name" : USED IN vasp.converge ONLY. Name used in the .../config/calctype.calc/NAME/property_i directory scheme, where, if not specified, "prop"_converge is used as NAME
    """
    try:
        with open(filename, 'rb') as file:
            settings = json.loads(file.read().decode('utf-8'))
    except (IOError, ValueError) as e:
        print("Error reading settings file:", filename)
        raise e

    required = ["queue", "ppn", "walltime"]

    select_one = [["nodes", "atom_per_proc", "nodes_per_image"]]

    optional = [
        "account", "pmem", "priority", "constraint", "message", "email", "qos",
        "npar", "ncore", "kpar", "ncpus", "vasp_cmd", "run_limit",
        "nrg_convergence", "encut", "kpoints", "extra_input_files", "move",
        "copy", "remove", "compress", "backup", "initial", "final",
        "strict_kpoints", "err_types", "preamble", "prerun", "postrun", "prop",
        "prop_start", "prop_stop", "prop_step", "tol", "tol_amount", "name",
        "fine_ngx", "CI_neb", "n_images", "software", "method",
        "endstate_calctype", "initial_deformation"
    ]

    for key in required:
        if not key in settings:
            raise VaspWrapperError(key + "' missing from: '" + filename + "'")

    if len(select_one):
        for key_list in select_one:
            if not [key in settings for key in key_list].count(True) == 1:
                raise VaspWrapperError(
                    "Declare one and only of the following options: '" +
                    "' or '".join(key_list) + "' in file: '" + filename + "'")
            for key in key_list:
                if not key in settings:
                    settings[key] = None

    for key in optional:
        if not key in settings:
            if key.lower() in [
                    "extra_input_files", "remove", "compress", "backup"
            ]:
                settings[key] = []
            elif key.lower() in ["move"]:
                settings[key] = casm.vasp.io.DEFAULT_VASP_MOVE_LIST
            elif key.lower() in ["copy"]:
                settings[key] = casm.vasp.io.DEFAULT_VASP_COPY_LIST
            # elif key.lower() in ["remove"]:
            #     settings[key] = casm.vasp.io.DEFAULT_VASP_REMOVE_LIST
            else:
                settings[key] = None

    if type(settings["remove"]) == list:
        if 'default' in settings["remove"]:
            settings["remove"] += casm.vasp.io.DEFAULT_VASP_REMOVE_LIST
    elif type(settings["remove"]) == str:
        if settings["remove"].lower() == 'default':
            settings["remove"] = casm.vasp.io.DEFAULT_VASP_REMOVE_LIST
        else:
            settings["remove"] = [settings["remove"]]
    if settings["priority"] == None:
        settings["priority"] = 0
    if settings["extra_input_files"] == None:
        settings["extra_input_files"] = []
    if settings["strict_kpoints"] == None:
        settings["strict_kpoints"] = False
    if settings["fine_ngx"] == None:
        settings["fine_ngx"] = False
    for k in settings.keys():
        if k not in required + optional + [
                key for key_list in select_one for key in key_list
        ]:
            raise VaspWrapperError("unknown key '" + k + "' found in: '" +
                                   filename + "'")

    return settings


def write_settings(settings, filename):
    """ Write 'settings' as json file, 'filename' """
    with open(filename, 'wb') as file:
        file.write(six.u(json.dump(settings, file, indent=4)).encode('utf-8'))


def vasp_input_file_names(dir, configname, clex, calc_subdir="", is_neb=False):
    """
    Collect casm.vaspwrapper input files from the CASM project hierarchy

    Looks for:

      INCAR:
        The base INCAR file used for calculations. Found via:
          DirectoryStructure.settings_path_crawl

      KPOINTS:
        The KPOINTS file specifying the k-point grid for a reference structure
        which is then scaled to be approximately the same density for other
        structures. Found via:
          DirectoryStructure.settings_path_crawl

      KPOINTS_REF: (optional)
        The reference structure used to determine the k-point density, if not
        running in Auto mode. If running VASP with AUTO KPOINTS mode, this file
        is not necessary. Found via:
          DirectoryStructure.settings_path_crawl

      structurefile:
        The CASM structure.json or VASP POSCAR file giving the initial structure to be calculated.

      SPECIES:
        The SPECIES file specifying Vasp settings for each species in the structure.


    Arguments
    ---------

      dir: casm.project.DirectoryStructure instance
        CASM project directory hierarchy

      configname: str
        The name of the configuration to be calculated

      clex: casm.project.ClexDescription instance
        The cluster expansion being worked on. Used for the 'calctype' settings.


    Returns
    -------

      filepaths: tuple(INCAR, KPOINTS, KPOINTS_REF, structurefile, SPECIES)
        A tuple containing the paths to the vaspwrapper input files


    Raises
    ------
      If any required file is not found.

    """
    # Find required input files in CASM project directory tree
    incarfile = dir.settings_path_crawl("INCAR", configname, clex)
    ref_kpointsfile = dir.settings_path_crawl("KPOINTS", configname, clex)
    ref_structurefile = dir.settings_path_crawl("POSCAR", configname, clex)
    structurefile = dir.structure_json(configname, calc_subdir)
    speciesfile = dir.settings_path_crawl("SPECIES", configname, clex)

    # Verify that required input files exist
    if incarfile is None:
        raise vasp.VaspError(
            "vasp_input_file_names failed. No INCAR file found in CASM project."
        )
    if ref_kpointsfile is None:
        raise vasp.VaspError(
            "vasp_input_file_names failed. No KPOINTS file found in CASM project."
        )
    if ref_structurefile is None:
        warnings.warn(
            "No reference POSCAR file found in CASM project. I hope your KPOINTS mode is A/AUTO/Automatic or this will fail!",
            vasp.VaspWarning)
    if structurefile is None and not is_neb:
        raise vasp.VaspError(
            "vasp_input_file_names failed. No structure.json file found for this configuration."
        )
    if speciesfile is None:
        raise vasp.VaspError(
            "vasp_input_file_names failed. No SPECIES file found in CASM project."
        )

    return (incarfile, ref_kpointsfile, ref_structurefile, structurefile,
            speciesfile)


def read_properties(filename):
    """ Read a properties.calc.json"""
    required = [
        "atom_type", "atoms_per_type", "coord_mode", "relaxed_basis",
        "relaxed_energy", "relaxed_forces", "relaxed_lattice"
    ]
    optional = ["relaxed_magmom", "relaxed_mag_basis"]

    with open(filename, 'rb') as myfile:
        properties = json.loads(myfile.read().decode('utf-8'))

    for key in required:
        if not key in properties:
            raise VaspWrapperError(key + "' missing from: '" + filename + "'")

    for key in optional:
        if not key in properties:
            properties[key] = None

    return properties



""" Script for running vasp convergences """
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import argparse
import casm.vaspwrapper


def main():
    """ Execute vasp.converge """
    parser = argparse.ArgumentParser(
        description='''Set up, run, and collect data on a VASP Convergence over
                       a specified property.''')

    parser.add_argument('configdir',
                        help='A path to a configuration in a CASM project')
    parser.add_argument(
        '-c',
        '--collect',
        action='store_true',
        help='Switch to collection mode instead of submission mode (default)')

    args = parser.parse_args()

    print("Begin vasp.converge")

    if args.collect:
        convergence = casm.vaspwrapper.Converge(args.configdir)

        convergence.collect()

        print("Finish vasp.converge\n\n")
    else:

        convergence = casm.vaspwrapper.Converge(args.configdir)

        convergence.submit()

        print("Finish vasp.converge\n\n")


if __name__ == '__main__':
    main()



"""casm.project file io"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import json
import six
from casm.misc import compat, noindent


def write_eci(proj, eci, fit_details=None, clex=None, verbose=False):
    """Write eci.json

    Arguments
    ---------

    proj: casm.project.Project
        The CASM project

    eci: List[(index, value)]
        List of tuple containing basis function index and coefficient value:

            index (int): linear index of basis function
            value (float): ECI value

    fit_details: Dict
        Description of the fitting method used to generate the ECI, usually as output by `casm.learn.to_json`. It is added to the eci.json file under the attribute `"fit"`.

    clex: casm.project.ClexDescription
        Specifies where to write the ECI. Default value is  `proj.settings.default_clex`.

    """
    dir = proj.dir
    if clex is None:
        clex = proj.settings.default_clex

    # read basis.json
    filename = dir.basis(clex)
    with open(filename, 'r') as f:
        j = json.load(f)

    # edit to add fitting settings
    j["fit"] = fit_details

    # make linear_function_index: (linear_orbit_index, cluster_function_index)
    function_keys = {}
    try:
        linear_orbit_index = 0
        for orbit in j["orbits"]:
            cluster_function_index = 0
            for cluster_function in orbit["cluster_functions"]:
                linear_function_index = cluster_function["linear_function_index"]
                function_keys[linear_function_index] = (linear_orbit_index, cluster_function_index)
                cluster_function_index += 1
            linear_orbit_index += 1
    except:
        raise Exception("Formatting error in basis.json")

    # edit to add eci
    for index, value in eci:
        linear_orbit_index = function_keys[index][0]
        cluster_function_index = function_keys[index][1]
        orbit = j["orbits"][linear_orbit_index]
        cluster_function = orbit["cluster_functions"][cluster_function_index]
        if cluster_function["linear_function_index"] != index:
            msg = "Formatting error in '" + str(filename) + "': "
            msg += " linear_function_index mismatch"
            raise Exception(msg)
        cluster_function["eci"] = value

    # pretty printing -- could be improved
    cspecs_params = j["bspecs"]["cluster_specs"]["params"]
    if "generating_group" in cspecs_params:
        cspecs_params["generating_group"] = noindent.NoIndent(
            cspecs_params["generating_group"])
    for entry in j["orbits"]:
        sites = entry["prototype"]["sites"]
        for i in range(len(sites)):
            sites[i] = noindent.NoIndent(sites[i])
    for site in j["prim"]["basis"]:
        site["coordinate"] = noindent.NoIndent(site["coordinate"])
        site["occupants"] = noindent.NoIndent(site["occupants"])
    for i in range(len(j["prim"]["lattice_vectors"])):
        j["prim"]["lattice_vectors"][i] = noindent.NoIndent(
            j["prim"]["lattice_vectors"][i])

    # write eci.json
    filename = dir.eci(clex)

    if verbose:
        print("Writing:", filename, "\n")
    with open(filename, 'w') as f:
        f.write(json.dumps(j, indent=2, cls=noindent.NoIndentEncoder))

    # refresh proj to reflect new eci
    proj.refresh(clear_clex=True)
